/* ARCADEmini USB driver for Windows
 * version 1.2.0 date 2005-06-06
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003-2005 Stefan Schuermans <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <wdm.h>
#include <stdarg.h>
#include <stdio.h>
#include <usbdi.h>
#include <usbdlib.h>

#include "AmUsbInternal.h"

//GUID: {B57A5F08-C571-8909-8070-557E2741687A}
static const GUID AMUSB_GUID = { 0xB57A5f08, 0xC571, 0x8909, { 0x80, 0x70, 0x55, 0x7E, 0x27, 0x41, 0x68, 0x7A } };

//We keep a pending IO count ( extension->PendingIoCount )  in the device extension.
//The first increment of this count is done on adding the device.
//Subsequently, the count is incremented for each new IRP received and
//decremented when each IRP is completed or passed on.
//Transition to 'one' therefore indicates no IO is pending and signals
//deviceExtension->NoPendingIoEvent. This is needed for processing
//IRP_MN_QUERY_REMOVE_DEVICE
//Transition to 'zero' signals an event ( extension->RemoveEvent )
//to enable device removal. This is used in processing for IRP_MN_REMOVE_DEVICE

//decrement the IO counter
//pDeviceObject: pointer to the functional device object
//return value: the new IO count
LONG AmUsbDecIoCount( IN PDEVICE_OBJECT pDeviceObject )
{
	PDEVICE_EXTENSION pDeviceExtension;
	LONG IoCount;
	KIRQL OldIrql;

	//get pointer to the device extension
	pDeviceExtension = pDeviceObject->DeviceExtension;

	//lock IO count
	KeAcquireSpinLock( &pDeviceExtension->IoCountSpinLock, &OldIrql );

	//decrement the IO count
	IoCount = InterlockedDecrement( &pDeviceExtension->PendingIoCount );
	//we may never reach a value below 0
	AmUsbTrapCond( DBGLVL_HIGH, IoCount < 0 );

	//no pending IO
	if( IoCount == 1 )
		KeSetEvent( &pDeviceExtension->evNoPendingIo, 1, FALSE );

	//remove-device event
	if( IoCount == 0 )
		KeSetEvent( &pDeviceExtension->evRemove, 1, FALSE );

	//unlock IO count
	KeReleaseSpinLock( &pDeviceExtension->IoCountSpinLock, OldIrql );

	//debug message
	AmUsbDebug( DBGLVL_HIGH, ( "AmUsbDecIoCount(), pending io count = %d\r\n", IoCount ) );

	//return new IO count
	return IoCount;
}

//increment the IO counter
//pDeviceObject: pointer to the functional device object
//return value: the new IO count
LONG AmUsbIncIoCount( IN PDEVICE_OBJECT pDeviceObject )
{
	PDEVICE_EXTENSION pDeviceExtension;
	LONG IoCount;
	KIRQL OldIrql;

	//get pointer to the device extension
	pDeviceExtension = pDeviceObject->DeviceExtension;

	//lock IO count
	KeAcquireSpinLock( &pDeviceExtension->IoCountSpinLock, &OldIrql );

	//decrement the IO count
	IoCount = InterlockedIncrement( &pDeviceExtension->PendingIoCount );

	//unlock IO count
	KeReleaseSpinLock( &pDeviceExtension->IoCountSpinLock, OldIrql );

	//debug message
	AmUsbDebug( DBGLVL_HIGH, ( "AmUsbIncIoCount(), pending io count = %d\r\n", IoCount ) );

	//return new IO count
	return IoCount;
}

//process a system control IRP
//pDeviceObject: pointer to the device object
//pIrp: pointer to the IRP to process
//return value: status returned from lower driver
NTSTATUS AmUsbSysControl( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp )
{
	PIO_STACK_LOCATION pIrpStack;
	PDEVICE_EXTENSION pDeviceExtension;
	NTSTATUS Status = STATUS_SUCCESS;
	PDEVICE_OBJECT pStackDeviceObject;

	//no error yet
	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;

	//get pointer to current location in IRP
	pIrpStack = IoGetCurrentIrpStackLocation( pIrp );

	//get pointer to device extension
	pDeviceExtension = pDeviceObject->DeviceExtension;
	pStackDeviceObject = pDeviceExtension->pTopOfStackDeviceObject;

	//debug message
    AmUsbDebug( DBGLVL_HIGH, ( "entering AmUsbSysControl()\r\n" ) );

	//increment IO counter, because we are calling another driver
	AmUsbIncIoCount( pDeviceObject );

	//major function must be system control
	AmUsbAssert( pIrpStack->MajorFunction == IRP_MJ_SYSTEM_CONTROL );

	//pass the IRP to the lower driver
	IoCopyCurrentIrpStackLocationToNext( pIrp );
	Status = IoCallDriver( pStackDeviceObject, pIrp );

	//decrement IO counter, because we returned from call to other driver
	AmUsbDecIoCount( pDeviceObject );

	//debug message
    AmUsbDebug( DBGLVL_HIGH, ( "exiting AmUsbSysControl() with status 0x%x\r\n", Status ) );

	return Status;
}

//driver is unloaded, free all resources, etc.
//pDriverObject: pointer to the driver object
VOID AmUsbUnload( IN PDRIVER_OBJECT pDriverObject )
{
	//debug message
    AmUsbDebug( DBGLVL_HIGH, ( "entering AmUsbUnload\r\n" ) );

	//nothing to do here, because we did not allocate anything in DriverEntry()

	//debug message
    AmUsbDebug( DBGLVL_HIGH, ( "exiting AmUsbUnload\r\n" ) );
}

//create and initialize a GUID-based symbolic link to the dvice to be used from user-space
//pDeviceObject: pointer to the device object
//pDeviceLinkUnicodeString: pointer to a unicode string structure allocated by the caller
//  if this function is successfull, it initializes the string, allocates the buffer
//  and puts kernel-mode path to the symbolic link tos this device into it
//return value: STATUS_SUCCESS / STATUS_UNSECCESSFUL
NTSTATUS AmUsbSymbolicLink( IN PDEVICE_OBJECT pDeviceObject, IN OUT PUNICODE_STRING pDeviceLinkUnicodeString )
{
	NTSTATUS Status = STATUS_SUCCESS;

	//create new device interface and symbolic link to it
	Status = IoRegisterDeviceInterface( pDeviceObject, &AMUSB_GUID, NULL, pDeviceLinkUnicodeString );

	//debug message, in case of error
    AmUsbDebugCond( DBGLVL_MEDIUM, ( !NT_SUCCESS( Status ) ), ( "IoRegisterDeviceInterface() FAILED\r\n" ) );

	//success
	if( NT_SUCCESS( Status ) )
	{
		//enable new device interface
		Status = IoSetDeviceInterfaceState( pDeviceLinkUnicodeString, TRUE );

		//debug message, in case of error
		AmUsbDebugCond( DBGLVL_MEDIUM, ( !NT_SUCCESS( Status ) ), ( "SetDeviceInterfaceState() FAILED\r\n" ) );
	}

	return Status;
}

//creates a device object
//pDriverObject: pointer to the driver object
//pPhysicalDeviceObject: pointer to the physical device object
//ppDeviceObject: pointer to location where to put the pointer to the new device object
//return value: STATUS_SUCCESS / STATUS_UNSECCESSFUL
NTSTATUS AmUsbCreateDeviceObject( IN PDRIVER_OBJECT pDriverObject, IN PDEVICE_OBJECT pPhysicalDeviceObject, IN PDEVICE_OBJECT * ppDeviceObject )
{
	NTSTATUS Status;
	UNICODE_STRING DeviceLinkUnicodeString;
	PDEVICE_EXTENSION pDeviceExtension;
	int SymLinkNo;
	char SymLinkName[256];
	ANSI_STRING SymLinkAnsiString;

	//debug message
    AmUsbDebug( DBGLVL_DEFAULT, ( "entering AmUsbCreateDeviceObject()\r\n" ) );

	//create a device interface, and a symbolic link to it
	Status = AmUsbSymbolicLink( pPhysicalDeviceObject, &DeviceLinkUnicodeString );

	//debug message, in case of success
    AmUsbDebugCond( DBGLVL_DEFAULT, NT_SUCCESS( Status ),
			( "created GUID_CLASS_AMUSB-based device name\r\n  %ws\r\n  length %d, max. length %d\r\n",
			DeviceLinkUnicodeString.Buffer,
			DeviceLinkUnicodeString.Length,
			DeviceLinkUnicodeString.MaximumLength ) );
	//debug message, in case of error
    AmUsbDebugCond( DBGLVL_DEFAULT, !NT_SUCCESS( Status ),
			( "FAILED to create GUID_CLASS_AMUSB-based device name\r\n" ) );

	//success
	if( NT_SUCCESS( Status ) )
	{
		//create the device
		Status = IoCreateDevice( pDriverObject,
				sizeof( DEVICE_EXTENSION ),	NULL,
				FILE_DEVICE_UNKNOWN, FILE_AUTOGENERATED_DEVICE_NAME,
				FALSE, ppDeviceObject );

		//success
		if( NT_SUCCESS( Status ) )
		{
			//get pointer to device extension
			pDeviceExtension = (PDEVICE_EXTENSION)((*ppDeviceObject)->DeviceExtension);
		}
		else
		{
			//debug messages
			AmUsbDebug( DBGLVL_DEFAULT, ( "IoCreateDevice() FAILED\r\n" ) );
			AmUsbDebug( DBGLVL_DEFAULT, ( "exiting AmUsbCreateDeviceObject()\r\n" ) );
			//exit function
			return Status;
		}

		//zero memory of device extension
		memset( pDeviceExtension, 0, sizeof( DEVICE_EXTENSION ) );

		//set a symbolic link
		//default: no symlink set
		pDeviceExtension->SymLinkSet = FALSE;
		//try all numbers from 0 to (AmUsbSymLinkNoMax-1) behind symlink
		for( SymLinkNo = 0; SymLinkNo < AmUsbSymLinkNoMax; SymLinkNo++ )
		{
			//assemble name of symlink to try
			_snprintf( SymLinkName, sizeof( SymLinkName ), "%s%d", AmUsbSymLinkNamePrefix, SymLinkNo );
			RtlInitAnsiString( &SymLinkAnsiString, SymLinkName );
			//convert into unicode string
			pDeviceExtension->SymLink.Buffer = NULL;
			if( RtlAnsiStringToUnicodeString( &pDeviceExtension->SymLink, &SymLinkAnsiString, TRUE ) == STATUS_SUCCESS )
			{
				//try to set symlink
				if( IoCreateSymbolicLink( &pDeviceExtension->SymLink, &DeviceLinkUnicodeString ) == STATUS_SUCCESS )
				{
					//symlink is set
					pDeviceExtension->SymLinkSet = TRUE;
					//debug message
					AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbSymbolicLink: added symlink \"%s\"\r\n", SymLinkName ) );
					//end loop
					break;
				}
			}
			//free unicode string
			RtlFreeUnicodeString( &pDeviceExtension->SymLink );
		}

		//remember the name of out device
		RtlCopyMemory( pDeviceExtension->DeviceLinkNameBuffer,
				DeviceLinkUnicodeString.Buffer, DeviceLinkUnicodeString.Length );

		//initialize events
		//triggered when there is no pending io and device is removed
		KeInitializeEvent( &pDeviceExtension->evRemove, NotificationEvent, FALSE );
		//triggered when self-requested power IRPs complete
		KeInitializeEvent( &pDeviceExtension->evSelfRequestedPowerIrp, NotificationEvent, FALSE );

		//triggered when there is no pending io (pending io count == 1)
		KeInitializeEvent( &pDeviceExtension->evNoPendingIo, NotificationEvent, FALSE );

		//initialize spinlock used to protect inc/dec iocount logic
		KeInitializeSpinLock( &pDeviceExtension->IoCountSpinLock );

		//free buffer of unicode string used to init interface
		RtlFreeUnicodeString( &DeviceLinkUnicodeString );
	}

	//debug message
    AmUsbDebug( DBGLVL_DEFAULT, ( "exiting AmUsbCreateDeviceObject()\r\n" ) );

	return Status;
}

//pass a URB to the USBD class driver
//pDeviceObject: pointer to the physical device object
//pUrb: pointer to an already formatted URB
//return value: STATUS_SUCCESS / STATUS_UNSECCESSFUL
NTSTATUS AmUsbCallUsbd( IN PDEVICE_OBJECT pDeviceObject, IN PURB pUrb )
{
	NTSTATUS NtStatus, Status = STATUS_SUCCESS;
	PDEVICE_EXTENSION pDeviceExtension;
	PIRP pIrp;
	KEVENT Event;
	IO_STATUS_BLOCK IoStatus;
	PIO_STACK_LOCATION pNextStack;

	//debug message
    AmUsbDebug( DBGLVL_MAXIMUM, ( "entering AmUsbCallUsbd\r\n" ) );

	//get pointer to device extension
	pDeviceExtension = pDeviceObject->DeviceExtension;

	//issue a synchronous request
	KeInitializeEvent( &Event, NotificationEvent, FALSE );
	pIrp = IoBuildDeviceIoControlRequest(
			IOCTL_INTERNAL_USB_SUBMIT_URB,
			pDeviceExtension->pTopOfStackDeviceObject, //points to the next-lower driver's device object
			NULL, //no optional input buffer
			0, //no input buffer len
			NULL, //no optional output bufer
			0, //no output buffer len
			TRUE, //call target driver's dispatch function for IRP_MJ_INTERNAL_DEVICE_CONTROL
			&Event, //event to be signalled on completion
			&IoStatus );  //IO status block to be set when the request is completed

	//call the class driver to perform the operation
	pNextStack = IoGetNextIrpStackLocation( pIrp ); //get next driver
	AmUsbAssert( pNextStack != NULL ); //check if there is one driver
	pNextStack->Parameters.Others.Argument1 = pUrb; //pass URB to USB driver stack
	NtStatus = IoCallDriver( pDeviceExtension->pTopOfStackDeviceObject, pIrp ); //call class driver

	//debug message
    AmUsbDebug( DBGLVL_MAXIMUM, ( "IoCallDriver USBD returned 0x%x\r\n", NtStatus ) );

	//if the returned status is PENDING, wait for the request to complete
	if( NtStatus == STATUS_PENDING )
		Status = KeWaitForSingleObject( &Event, Suspended, KernelMode, FALSE, NULL );
	//if the returned status id not PENDING, so set the IO status to the returned status
	else
		IoStatus.Status = NtStatus;

	//debug message
    AmUsbDebug( DBGLVL_MAXIMUM, ( "URB status = 0x%x, status = 0x%x, IRP status 0x%x\r\n",
			pUrb->UrbHeader.Status, Status, IoStatus.Status ) );

	//get error code that USBD put into IoStatus
	NtStatus = IoStatus.Status;

	//debug message, in case of error
    AmUsbDebugCond( DBGLVL_MAXIMUM, !NT_SUCCESS( NtStatus ),
			( "AmUsbCallUsbd FAILED with status 0x%x\r\n", NtStatus ) );

	return NtStatus;
}

//pDeviceObject: pointer to the device object
//pConfigDescr: pointer to the USB config descriptor including interface and enpoints descriptors
//return value: NT status code
NTSTATUS AmUsbSelectInterface( IN PDEVICE_OBJECT pDeviceObject, IN PUSB_CONFIGURATION_DESCRIPTOR pConfigDescr )
{
	PDEVICE_EXTENSION pDeviceExtension;
	NTSTATUS Status;
	PURB pUrb = NULL;
	ULONG i;
	PUSB_INTERFACE_DESCRIPTOR pInterfaceDescriptor = NULL;
	PUSBD_INTERFACE_INFORMATION pInterface = NULL;
	USHORT Siz;

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "entering AmUsbSelectInterface\r\n" ) );

	//get pointer to device extension
	pDeviceExtension = pDeviceObject->DeviceExtension;

	//create a set config URB
	pUrb = USBD_CreateConfigurationRequest( pConfigDescr, &Siz );
	//we've got the URB
	if( pUrb )
	{
		//get the pointer to an (the one and only) interface
		pInterfaceDescriptor = USBD_ParseConfigurationDescriptorEx(
				pConfigDescr, pConfigDescr, //search from start of config descriptor
				-1, -1, -1, -1, -1 ); //don't care for interface, alternate setting, class, subclass or protocol
		//we've not got the interface
		if( !pInterfaceDescriptor )
		{
			//debug message
			AmUsbDebug( DBGLVL_HIGH, ( "USBD_ParseConfigurationDescriptorEx() FAILED\r\n" ) );
			//free URB
			ExFreePool( pUrb );
			return STATUS_INSUFFICIENT_RESOURCES;
		}
		//get pointer to interface
		pInterface = &pUrb->UrbSelectConfiguration.Interface;

		//initialize pipes
		for( i = 0; i < pInterface->NumberOfPipes; i++ )
		{
			pInterface->Pipes[i].MaximumTransferSize = AmUsbMaximumTransferSize;
			pInterface->Pipes[i].PipeFlags = 0;
		}

		//build and execute the set config URB
		UsbBuildSelectConfigurationRequest( pUrb, Siz, pConfigDescr );
		Status = AmUsbCallUsbd( pDeviceObject, pUrb );
	} //if( pUrb )
	//we've not got the URB
	else
	{
		//debug message
		AmUsbDebug( DBGLVL_HIGH, ( "USBD_CreateConfigurationRequest() FAILED\r\n" ) );
		//set status to indicate error
		Status = STATUS_INSUFFICIENT_RESOURCES;
	} //if( pUrb ) ... else

	//everthing worked - we've got a configured USB device
	if( NT_SUCCESS( Status ) )
	{
		//save configuration handle
		pDeviceExtension->hUsbConfiguration = pUrb->UrbSelectConfiguration.ConfigurationHandle;

		//allocate memory for a copy of the interface
		pDeviceExtension->pUsbInterface = ExAllocatePool( NonPagedPool, pInterface->Length );
		//we've got the memory
		if( pDeviceExtension->pUsbInterface )
		{
			ULONG j;

			//save a copy of the interface information returned
			RtlCopyMemory( pDeviceExtension->pUsbInterface, pInterface, pInterface->Length );

			//dump the interface to debug out
			AmUsbDebug( DBGLVL_MEDIUM,( "---------\r\n" ) );
			AmUsbDebug( DBGLVL_MEDIUM,( "# pipes 0x%x\r\n", pDeviceExtension->pUsbInterface->NumberOfPipes ) );
			AmUsbDebug( DBGLVL_MEDIUM,( "length 0x%x\r\n", pDeviceExtension->pUsbInterface->Length ) );
			AmUsbDebug( DBGLVL_MEDIUM,( "alt setting 0x%x\r\n", pDeviceExtension->pUsbInterface->AlternateSetting ) );
			AmUsbDebug( DBGLVL_MEDIUM,( "interface number 0x%x\r\n", pDeviceExtension->pUsbInterface->InterfaceNumber ) );
			AmUsbDebug( DBGLVL_MEDIUM,( "class, subclass, protocol 0x%x 0x%x 0x%x\r\n",
					pDeviceExtension->pUsbInterface->Class,
					pDeviceExtension->pUsbInterface->SubClass,
					pDeviceExtension->pUsbInterface->Protocol ) );
			//dump the pipe info to debug out
			for( j=0; j < pInterface->NumberOfPipes; j++ )
			{
				PUSBD_PIPE_INFORMATION pPipeInformation;

				pPipeInformation = &pDeviceExtension->pUsbInterface->Pipes[j];

				AmUsbDebug( DBGLVL_MEDIUM, ( "---------\r\n" ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "PipeType 0x%x\r\n", pPipeInformation->PipeType ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "EndpointAddress 0x%x\r\n", pPipeInformation->EndpointAddress ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "MaxPacketSize 0x%x\r\n", pPipeInformation->MaximumPacketSize ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "Interval 0x%x\r\n", pPipeInformation->Interval ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "Handle 0x%x\r\n", pPipeInformation->PipeHandle ) );
				AmUsbDebug( DBGLVL_MEDIUM, ( "MaximumTransferSize 0x%x\r\n", pPipeInformation->MaximumTransferSize ) );
			}
			AmUsbDebug( DBGLVL_MEDIUM, ( "---------\r\n" ) );
		} //if( pDeviceExtension->UsbInterface )
	} //if( NT_SUCCESS( Status ) )

	//free the URB, if we allocated one
	if( pUrb )
		ExFreePool( pUrb );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "exiting AmUsbSelectInterface with status 0x%x\r\n", Status ) );

	return Status;
}

//initialize a device on USB, select and save configuration
//pDeviceObject: pointer to device to initialize
//return value: NT status code
NTSTATUS AmUsbConfigureDevice( IN PDEVICE_OBJECT pDeviceObject )
{
	PDEVICE_EXTENSION pDeviceExtension;
	NTSTATUS Status;
	PURB pUrb;
	ULONG Siz;

	//debug message
    AmUsbDebug( DBGLVL_HIGH, ( "entering AmUsbConfigureDevice\r\n" ) );

	//get pointer to device extension
	pDeviceExtension = pDeviceObject->DeviceExtension;
	//check that config descriptor is not yet there
	AmUsbAssert( pDeviceExtension->pUsbConfigurationDescriptor == NULL );

	//allocate URB
	pUrb = ExAllocatePool( NonPagedPool, sizeof( struct _URB_CONTROL_DESCRIPTOR_REQUEST ) );
	if( !pUrb )
		return STATUS_INSUFFICIENT_RESOURCES;

	//when USB_CONFIGURATION_DESCRIPTOR_TYPE is specified for DescriptorType
	//in a call to UsbBuildGetDescriptorRequest(),
	//all interface, endpoint, class-specific, and vendor-specific descriptors 
	//for the configuration also are retrieved. 
	//The caller must allocate a buffer large enough to hold all of this 
	//information or the data is truncated without error.
	//therefore the 'Siz' set below is just a 'good guess', and we may have to retry
	Siz = sizeof( USB_CONFIGURATION_DESCRIPTOR ) + 512;

	//we will break out of this 'retry loop' when UsbBuildGetDescriptorRequest()
	//has a big enough deviceExtension->UsbConfigurationDescriptor buffer not to truncate
	while( 1 )
	{
		//allocate memory block for the descriptors
		pDeviceExtension->pUsbConfigurationDescriptor = ExAllocatePool( NonPagedPool, Siz );
		//no memory available
		if( !pDeviceExtension->pUsbConfigurationDescriptor )
		{
			ExFreePool( pUrb );
			return STATUS_INSUFFICIENT_RESOURCES;
		}

		//build a get descriptor request
		UsbBuildGetDescriptorRequest( pUrb, sizeof (struct _URB_CONTROL_DESCRIPTOR_REQUEST),
				USB_CONFIGURATION_DESCRIPTOR_TYPE, 0, 0,
				pDeviceExtension->pUsbConfigurationDescriptor, NULL, Siz, NULL );
		//execute the request
		Status = AmUsbCallUsbd( pDeviceObject, pUrb );

		//debug message
	    AmUsbDebug( DBGLVL_HIGH, ( "Configuration Descriptor = 0x%x, len 0x%x\r\n",
				pDeviceExtension->pUsbConfigurationDescriptor,
				pUrb->UrbControlDescriptorRequest.TransferBufferLength ) );

		//if we got some data see if it was enough
		//note: we may get an error in URB because of buffer overrun
		if( pUrb->UrbControlDescriptorRequest.TransferBufferLength > 0
				&& pDeviceExtension->pUsbConfigurationDescriptor->wTotalLength > Siz)
		{
			//get needed size and free old memory block
			Siz = pDeviceExtension->pUsbConfigurationDescriptor->wTotalLength;
			ExFreePool( pDeviceExtension->pUsbConfigurationDescriptor );
			pDeviceExtension->pUsbConfigurationDescriptor = NULL;
		}
		else
			//we got the entire descriptors
			break;
	} //while( 1 )

	//free URB
	ExFreePool( pUrb );

	//check, if we've got the descriptor
	AmUsbAssert( pDeviceExtension->pUsbConfigurationDescriptor != NULL );

	//we have the configuration descriptor for the configuration we want
	//now we issue the select configuration command to get
	//the pipes associated with this configuration
	Status = AmUsbSelectInterface( pDeviceObject, pDeviceExtension->pUsbConfigurationDescriptor );

	//debug message
	AmUsbDebug( DBGLVL_HIGH, ("exiting AmUsbConfigureDevice with status 0x%x\r\n", Status ) );

	return Status;
}

//reset a USB pipe
//pDeviceObject: pointer to the device object
//pPieInfo: pointer to the pipe information
//return value: NT status code
NTSTATUS AmUsbResetPipe( IN PDEVICE_OBJECT pDeviceObject, IN PUSBD_PIPE_INFORMATION pPipeInfo )
{
	NTSTATUS Status;
	PURB pUrb;
	PDEVICE_EXTENSION pDeviceExtension;

	//debug message
	AmUsbDebug( DBGLVL_DEFAULT, ( "entering AmUsbResetPipe 0x%x\r\n", pPipeInfo ) );

	//get pointer to device extension
	pDeviceExtension = pDeviceObject->DeviceExtension;

	//allocate memory for an URB
	pUrb = ExAllocatePool( NonPagedPool, sizeof( struct _URB_PIPE_REQUEST ) );
	//we've got the memory
	if( pUrb )
	{
		//set up URB for resetting the pipe
		pUrb->UrbHeader.Length = sizeof (struct _URB_PIPE_REQUEST);
		pUrb->UrbHeader.Function = URB_FUNCTION_RESET_PIPE;
		pUrb->UrbPipeRequest.PipeHandle = pPipeInfo->PipeHandle;
		//execute URB: reset pipe
		Status = AmUsbCallUsbd( pDeviceObject, pUrb );
		//free URB
		ExFreePool( pUrb );
	}
	//we've not got the memory
	else
		//set status to indicate the error
		Status = STATUS_INSUFFICIENT_RESOURCES;

	//debug message, in case of error
	AmUsbDebugCond( DBGLVL_DEFAULT, !NT_SUCCESS( Status ), ( "AmUsbResetPipe FAILED with 0x%x\r\n", Status ) );
	//debug message, in case of success
	AmUsbDebugCond( DBGLVL_DEFAULT, NT_SUCCESS( Status ), ( "AmUsbResetPipe succeeded with 0x%x\r\n", Status ) );

	//debug message
	AmUsbDebug( DBGLVL_DEFAULT, ( "exiting AmUsbResetPipe 0x%x\r\n", pPipeInfo ) );

	return Status;
}

//driver initialization entry point, called by the IO-system
//pDriverObject: pointer to the driver object
//pRegistryPath: pointer to unicode string with path to driver's registry-key
//return value: STATUS_SUCCESS / STATUS_UNSECCESSFUL
NTSTATUS DriverEntry( IN PDRIVER_OBJECT pDriverObject, IN PUNICODE_STRING pRegistryPath )
{
	NTSTATUS Status = STATUS_SUCCESS;

	//debug message
    AmUsbDebug( DBGLVL_HIGH, ( "entering DriverEntry(), RegistryPath=\r\n  %ws\r\n", pRegistryPath->Buffer ) );

	//set pointer to unload function
	pDriverObject->DriverUnload = AmUsbUnload;

	//set pointers to functions for PnP and power management
	pDriverObject->MajorFunction[IRP_MJ_SYSTEM_CONTROL] = AmUsbSysControl;
	pDriverObject->MajorFunction[IRP_MJ_PNP] = AmUsbPnp;
	pDriverObject->MajorFunction[IRP_MJ_POWER] = AmUsbPower;

	//set pointers to create, close, read, write and ioctl function
	pDriverObject->MajorFunction[IRP_MJ_CREATE] = AmUsbCreate;
	pDriverObject->MajorFunction[IRP_MJ_CLOSE] = AmUsbClose;
	pDriverObject->MajorFunction[IRP_MJ_READ] = AmUsbRead;
	pDriverObject->MajorFunction[IRP_MJ_WRITE] = AmUsbWrite;
	pDriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = AmUsbIoctl;

	//request PnP system to call us, when a device is pugged in
	pDriverObject->DriverExtension->AddDevice = AmUsbPnpAddDevice;

	//debug message
    AmUsbDebug( DBGLVL_HIGH, ( "exiting DriverEntry() with status 0x%x\r\n", Status ) );

	return Status;
}
