/* ARCADEmini USB driver for Windows
 * version 1.2.0 date 2005-06-06
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003-2005 Stefan Schuermans <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <wdm.h>
#include <stdarg.h>
#include <stdio.h>
#include <usbdi.h>
#include <usbdlib.h>

#include "AmUsbInternal.h"

//This routine is called when an IRP_MN_SET_POWER of type 'DevicePowerState'
//has been received by AmUsbPower().
//pDeviceObject: pointer to the device object for the class device
//DeviceState: device specific power state to set the device in to
//return value: for requests to DeviceState D0 ( fully on ), returns TRUE to signal caller 
//              that we must set a completion routine and finish there.
BOOLEAN AmUsbSetDevicePowerState( IN PDEVICE_OBJECT pDeviceObject, IN DEVICE_POWER_STATE DeviceState )
{
	NTSTATUS Status = STATUS_SUCCESS;
	PDEVICE_EXTENSION pDeviceExtension;
	BOOLEAN fRes = FALSE;

	pDeviceExtension = (PDEVICE_EXTENSION)pDeviceObject->DeviceExtension;

	switch( DeviceState )
	{
	case PowerDeviceD3:
		//Device will be going OFF

		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbSetDevicePowerState() PowerDeviceD3 (OFF)\r\n" ) );

		pDeviceExtension->CurrentDevicePowerState = DeviceState;
		break;

	case PowerDeviceD1:
	case PowerDeviceD2:
		//power states D1,D2 translate to USB suspend

		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbSetDevicePowerState() %s\r\n",
			AmUsbStringForDevState( DeviceState ) ) );

		pDeviceExtension->CurrentDevicePowerState = DeviceState;
		break;

	case PowerDeviceD0:
		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbSetDevicePowerState() PowerDeviceD0 (ON)\r\n" ) );
		//we'll need to finish the rest in the completion routine;
		//signal caller we're going to D0 and will need to set a completion routine
		fRes = TRUE;

		//caller will pass on to PDO (Physical Device Object)
		break;

	default:
		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbSetDevicePowerState() Bogus DeviceState = 0x%x\r\n", DeviceState ) );
	}

	return fRes;
}

//This is the completion routine set in a call to PoRequestPowerIrp()
//that was made in BulkUsb_ProcessPowerIrp() in response to receiving
//an IRP_MN_SET_POWER of type 'SystemPowerState' when the device was
//not in a compatible device power state. In this case, a pointer to
//the IRP_MN_SET_POWER Irp is saved into the FDO device extension 
//(pDeviceExtension->PowerIrp), and then a call must be
//made to PoRequestPowerIrp() to put the device into a proper power state,
//and this routine is set as the completion routine.
//We decrement our pending io count and pass the saved IRP_MN_SET_POWER Irp
//on to the next driver
//pDeviceObjectClass: pointer to the device object for the class device
//  NOTE: we must get our own device object from pContext
//pContext: driver defined context, in this case our own functional device object (FDO)
//return value: the final status from the operation
NTSTATUS AmUsbPoRequestCompletion( IN PDEVICE_OBJECT pDeviceObjectClass, IN UCHAR MinorFunction, IN POWER_STATE PowerState, IN PVOID pContext, IN PIO_STATUS_BLOCK pIoStatus )
{
	PIRP pIrp;
	PDEVICE_EXTENSION pDeviceExtension;
	PDEVICE_OBJECT pDeviceObject = (PDEVICE_OBJECT)pContext;
	NTSTATUS Status;

	pDeviceExtension = pDeviceObject->DeviceExtension;

	//Get the IRP we saved for later processing in AmUsbPower()
	//when we decided to request the Power IRP that this routine 
	//is the completion routine for.
	pIrp = pDeviceExtension->pPowerIrp;

	//We will return the status set by the PDO for the power request we're completing
	Status = pIoStatus->Status;

	AmUsbDebug( DBGLVL_HIGH, ( "entering AmUsbPoRequestCompletion()\r\n" ) );

	//we should not be in the midst of handling a self-generated power IRP
	AmUsbAssert( !pDeviceExtension->SelfPowerIrp );

	//we must pass down to the next driver in the stack
	IoCopyCurrentIrpStackLocationToNext( pIrp );

	//Calling PoStartNextPowerIrp() indicates that the driver is finished
	//with the previous power IRP, if any, and is ready to handle the next power IRP.
	//It must be called for every power IRP. Although power IRPs are completed only once,
	//typically by the lowest-level driver for a device, PoStartNextPowerIrp must be called
	//for every stack location. Drivers must call PoStartNextPowerIrp while the current IRP
	//stack location points to the current driver. Therefore, this routine must be called
	//before IoCompleteRequest, IoSkipCurrentStackLocation, and PoCallDriver.
	PoStartNextPowerIrp( pIrp );

	//PoCallDriver is used to pass any power IRPs to the PDO instead of IoCallDriver.
	//When passing a power IRP down to a lower-level driver, the caller should use
	//IoSkipCurrentIrpStackLocation or IoCopyCurrentIrpStackLocationToNext to copy the IRP to
	//the next stack location, then call PoCallDriver. Use IoCopyCurrentIrpStackLocationToNext
	//if processing the IRP requires setting a completion routine, or IoSkipCurrentStackLocation
	//if no completion routine is needed.

	PoCallDriver( pDeviceExtension->pTopOfStackDeviceObject, pIrp );

	AmUsbDecIoCount( pDeviceObject );

	AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPoRequestCompletion() exiting IRP_MN_SET_POWER\r\n" ) );

	pDeviceExtension->pPowerIrp = NULL;

	return Status;
}

//This routine is called when An IRP_MN_SET_POWER of type 'DevicePowerState'
//has been received by AmUsb_ProcessPowerIrp(), and that routine has  determined
// 1) the request is for full powerup ( to PowerDeviceD0 ), and
// 2) We are not already in that state
//A call is then made to PoRequestPowerIrp() with this routine set as the completion routine.
//pDeviceObject: pointer to the device object for the class device
//pIrp: Irp completed
//pContext: driver defined context.
//return value: final status from the operation
NTSTATUS AmUsbPowerIrpComplete( IN PDEVICE_OBJECT pNullDeviceObject, IN PIRP pIrp, IN PVOID pContext )
{
	NTSTATUS Status = STATUS_SUCCESS;
	PDEVICE_OBJECT pDeviceObject;
	PIO_STACK_LOCATION pIrpStack;
	PDEVICE_EXTENSION pDeviceExtension;

	AmUsbDebug( DBGLVL_HIGH, ( "entering AmUsbPowerIrpComplete\r\n" ) );

	//get device and device extension
	pDeviceObject = (PDEVICE_OBJECT)pContext;
	pDeviceExtension = (PDEVICE_EXTENSION)pDeviceObject->DeviceExtension;

	//If the lower driver returned PENDING, mark our stack location as pending also.
	if( pIrp->PendingReturned )
		IoMarkIrpPending( pIrp );

	pIrpStack = IoGetCurrentIrpStackLocation( pIrp );

	//We can assert that we're a device powerup-to D0 request,
	//because that was the only type of request we set a completion routine
	//for in the first place
	AmUsbAssert( pIrpStack->MajorFunction == IRP_MJ_POWER );
	AmUsbAssert( pIrpStack->MinorFunction == IRP_MN_SET_POWER );
	AmUsbAssert( pIrpStack->Parameters.Power.Type == DevicePowerState );
	AmUsbAssert( pIrpStack->Parameters.Power.State.DeviceState == PowerDeviceD0 );

	//Now that we know we've let the lower drivers do what was needed to power up,
	//we can set our device extension flags accordingly
	pDeviceExtension->CurrentDevicePowerState = PowerDeviceD0;

	pIrp->IoStatus.Status = Status;

	AmUsbDecIoCount( pDeviceObject );

	AmUsbDebug( DBGLVL_MEDIUM, ( "exiting AmUsbPowerIrpComplete(), IRP_MN_SET_POWER D0 done\r\n" ) );
	return Status;
}

//This is our FDO's dispatch table function for IRP_MJ_POWER.
//It processes the Power IRPs sent to the PDO for this device.
//For every power IRP, drivers must call PoStartNextPowerIrp and use PoCallDriver
//to pass the IRP all the way down the driver stack to the underlying PDO.
//pDeviceObject: pointer to our device object (FDO)
//pIrp: pointer to an I/O Request Packet
//return value: NT status code
NTSTATUS AmUsbPower( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp )
{
	PIO_STACK_LOCATION pIrpStack;
	NTSTATUS Status = STATUS_SUCCESS;
	PDEVICE_EXTENSION pDeviceExtension;
	BOOLEAN fGoingToD0 = FALSE;
	POWER_STATE SysPowerState, DesiredDevicePowerState;
	KEVENT evEvent;

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "entering AmUsbPower()\r\n" ) );

	//get device extension and IRP stack
	pDeviceExtension = pDeviceObject->DeviceExtension;
	pIrpStack = IoGetCurrentIrpStackLocation( pIrp );
	AmUsbIncIoCount( pDeviceObject );

	switch( pIrpStack->MinorFunction )
	{
	case IRP_MN_WAIT_WAKE:
		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPower() entering IRP_MN_WAIT_WAKE\r\n"));

		//A driver sends IRP_MN_WAIT_WAKE to indicate that the system should 
		//wait for its device to signal a wake event. The exact nature of the event
		//is device-dependent. 
		//Drivers send this IRP for two reasons: 
		//1) To allow a device to wake the system
		//2) To wake a device that has been put into a sleep state to save power
		//   but still must be able to communicate with its driver under certain circumstances. 
		//When a wake event occurs, the driver completes the IRP and returns 
		//STATUS_SUCCESS. If the device is sleeping when the event occurs, 
		//the driver must first wake up the device before completing the IRP. 
		//In a completion routine, the driver calls PoRequestPowerIrp to send a 
		//PowerDeviceD0 request. When the device has powered up, the driver can 
		//handle the IRP_MN_WAIT_WAKE request.

		//pDeviceExtension->DeviceCapabilities.DeviceWake specifies the lowest device power state (least powered)
		//from which the device can signal a wake event 
		pDeviceExtension->PowerDownLevel = pDeviceExtension->DeviceCapabilities.DeviceWake;
		if( (pDeviceExtension->CurrentDevicePowerState == PowerDeviceD0) ||
			(pDeviceExtension->DeviceCapabilities.DeviceWake > pDeviceExtension->CurrentDevicePowerState) )
		{
			//STATUS_INVALID_DEVICE_STATE is returned if the device in the PowerD0 state
			//or a state below which it can support waking, or if the SystemWake state
			//is below a state which can be supported. A pending IRP_MN_WAIT_WAKE will complete
			//with this error if the device's state is changed to be incompatible with the wake 
			//request.
			//If a driver fails this IRP, it should complete the IRP immediately without
			//passing the IRP to the next-lower driver.
			Status = STATUS_INVALID_DEVICE_STATE;
			pIrp->IoStatus.Status = Status;
			IoCompleteRequest( pIrp, IO_NO_INCREMENT );
			AmUsbDebug( DBGLVL_HIGH, ( "exiting AmUsbPower() with STATUS_INVALID_DEVICE_STATE\r\n" ) );
			AmUsbDecIoCount( pDeviceObject );
			return Status;
		}

		//flag we're enabled for wakeup
		pDeviceExtension->EnabledForWakeup = TRUE;

		//init an event for our completion routine to signal when PDO is done with this IRP
		KeInitializeEvent( &evEvent, NotificationEvent, FALSE );

		//If not failing outright, pass this on to our PDO for further handling
		IoCopyCurrentIrpStackLocationToNext( pIrp );

		//Set a completion routine so it can signal our event when
		//the PDO is done with the Irp
		IoSetCompletionRoutine( pIrp, AmUsbIrpCompletionRoutine,
							   &evEvent, //pass the event to the completion routine as the Context
							   TRUE, TRUE, TRUE ); //invoke on success, on error and on cancellation

		PoStartNextPowerIrp( pIrp );
		Status = PoCallDriver( pDeviceExtension->pTopOfStackDeviceObject, pIrp );

		//if PDO is not done yet, wait for the event to be set in our completion routine
		if( Status == STATUS_PENDING )
		{
			//wait for IRP to complete
			NTSTATUS WaitStatus = KeWaitForSingleObject( &evEvent,
					Suspended, KernelMode, FALSE, NULL );

			AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPower() done waiting for PDO to finish IRP_MN_WAIT_WAKE\r\n" ) );
		}

		//now tell the device to actually wake up
		AmUsbSelfSuspendOrActivate( pDeviceObject, FALSE );

		//flag we're done with wakeup IRP
		pDeviceExtension->EnabledForWakeup = FALSE;

		AmUsbDecIoCount( pDeviceObject );

		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPower() exiting IRP_MN_WAIT_WAKE\r\n" ) );
		break; //IRP_MN_WAIT_WAKE

	case IRP_MN_SET_POWER:
		{
		//The system power policy manager sends this IRP to set the system power state. 
		//A device power policy manager sends this IRP to set the device power state for a device.

		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPower() entering IRP_MN_SET_POWER\r\n" ) );

		//Set pIrp->IoStatus.Status to STATUS_SUCCESS to indicate that the device
		//has entered the requested state. Drivers cannot fail this IRP.

		switch( pIrpStack->Parameters.Power.Type )
		{
			case SystemPowerState:
				//get input system power state
				SysPowerState.SystemState = pIrpStack->Parameters.Power.State.SystemState;

				AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPower(), Set Power, type SystemPowerState = %s\r\n",
					AmUsbStringForSysState( SysPowerState.SystemState ) ) );

				//If system is in working state always set our device to D0
				//regardless of the wait state or system-to-device state power map
				if( SysPowerState.SystemState == PowerSystemWorking )
				{
					DesiredDevicePowerState.DeviceState = PowerDeviceD0;

					AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPower(), PowerSystemWorking, will set D0, not use state map\r\n" ) );
				}
				else
				{
					//set to corresponding system state if IRP_MN_WAIT_WAKE pending
					if( pDeviceExtension->EnabledForWakeup ) //got a WAIT_WAKE IRP pending?
					{
						//Find the device power state equivalent to the given system state.
						//We get this info from the DEVICE_CAPABILITIES struct in our device
						//extension (initialized in BulkUsb_PnPAddDevice())
						DesiredDevicePowerState.DeviceState =
							pDeviceExtension->DeviceCapabilities.DeviceState[SysPowerState.SystemState];

						AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPower(), IRP_MN_WAIT_WAKE pending, will use state map\r\n" ) );
					}
					else
					{  
						//if no wait pending and the system's not in working state, just turn off
						DesiredDevicePowerState.DeviceState = PowerDeviceD3;

						AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPower(), not EnabledForWakeup and the system's not in working state, settting PowerDeviceD3 (off)\r\n" ) );
					}
				}

				//We've determined the desired device state; are we already in this state?
				AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPower(), Set Power, DesiredDevicePowerState = %s\r\n",
					AmUsbStringForDevState( DesiredDevicePowerState.DeviceState ) ) );
				if( DesiredDevicePowerState.DeviceState != pDeviceExtension->CurrentDevicePowerState )
				{
					//No, request that we be put into this state
					//by requesting a new Power IRP from the Pnp manager
					pDeviceExtension->pPowerIrp = pIrp;
					Status = PoRequestPowerIrp( pDeviceExtension->pPhysicalDeviceObject,
							IRP_MN_SET_POWER, DesiredDevicePowerState,
							//completion routine will pass the IRP down to the PDO
							AmUsbPoRequestCompletion, pDeviceObject, NULL );
				}
				else
				{
					//Yes, just pass it on to PDO (Physical Device Object)
					IoCopyCurrentIrpStackLocationToNext( pIrp );
					PoStartNextPowerIrp( pIrp );
					Status = PoCallDriver( pDeviceExtension->pTopOfStackDeviceObject, pIrp );
					AmUsbDecIoCount( pDeviceObject );
					AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPower() exiting IRP_MN_SET_POWER\r\n" ) );
				}
				break;

			case DevicePowerState:

				AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPower, Set Power, type DevicePowerState = %s\r\n",
					AmUsbStringForDevState( pIrpStack->Parameters.Power.State.DeviceState ) ) );

				//For requests to D1, D2, or D3 (sleep or off states),
				//sets pDeviceExtension->CurrentDevicePowerState to DeviceState immediately.
				//This enables any code checking state to consider us as sleeping or off
				//already, as this will imminently become our state.

				//For requests to DeviceState D0 (fully on), sets fGoingToD0 flag TRUE
				//to flag that we must set a completion routine and update
				//pDeviceExtension->CurrentDevicePowerState there.
				//In the case of powering up to fully on, we really want to make sure
				//the process is completed before updating our CurrentDevicePowerState,
				//so no IO will be attempted or accepted before we're really ready.

				fGoingToD0 = AmUsbSetDevicePowerState( pDeviceObject, pIrpStack->Parameters.Power.State.DeviceState ); //returns TRUE for D0

				IoCopyCurrentIrpStackLocationToNext( pIrp );

				if( fGoingToD0 )
				{
					AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPower(), Set PowerIrp Completion Routine, fGoingToD0 = %d\r\n", fGoingToD0 ) );
					IoSetCompletionRoutine( pIrp, AmUsbPowerIrpComplete,
						   //Always pass FDO to completion routine as its Context;
						   //This is because the DriverObject passed by the system to the routine
						   //is the Physical Device Object (PDO) not the Functional Device Object (FDO)
						   pDeviceObject,
						   TRUE, TRUE, TRUE ); //invoke on success, on error and on cancellation of the IRP
				}

				PoStartNextPowerIrp( pIrp );
				Status = PoCallDriver( pDeviceExtension->pTopOfStackDeviceObject, pIrp );

				if( !fGoingToD0 ) //completion routine will decrement
					AmUsbDecIoCount( pDeviceObject );

				AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPower() exiting IRP_MN_SET_POWER\r\n" ) );
				break;
			} //switch( pIrpStack->Parameters.Power.Type )
		}
		break; //IRP_MN_SET_POWER

	case IRP_MN_QUERY_POWER:
		//A power policy manager sends this IRP to determine whether it can change
		//the system or device power state, typically to go to sleep.

		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPower() entering IRP_MN_QUERY_POWER\r\n" ) );

		//we do nothing special here, just let the PDO handle it
		IoCopyCurrentIrpStackLocationToNext( pIrp );
		PoStartNextPowerIrp( pIrp );
		Status = PoCallDriver( pDeviceExtension->pTopOfStackDeviceObject, pIrp );

		AmUsbDecIoCount( pDeviceObject );
		break; //IRP_MN_QUERY_POWER

	default:
		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbPower(), UNKNOWN POWER MESSAGE (%x)\r\n", pIrpStack->MinorFunction ) );
		//all unhandled power messages are passed on to the PDO
		IoCopyCurrentIrpStackLocationToNext( pIrp );
		PoStartNextPowerIrp( pIrp );
		Status = PoCallDriver( pDeviceExtension->pTopOfStackDeviceObject, pIrp );

		AmUsbDecIoCount( pDeviceObject );
	} //switch( pIrpStack->MinorFunction )

	AmUsbDebug( DBGLVL_MEDIUM, ( "exiting AmUsbPower() with status 0x%x\r\n", Status ) );
	return Status;
}

//This routine is called when the driver completes a self-originated power IRP 
//that was generated by a call to AmUsb_SelfSuspendOrActivate().
//We power down whenever the last pipe is closed and power up when the first pipe is opened.
//For power-up , we set an event in our FDO extension to signal this IRP done
//so the power request can be treated as a synchronous call.
//We need to know the device is powered up before opening the first pipe, for example.
//For power-down, we do not set the event, as no caller waits for powerdown complete.
//pDeviceObject: pointer to the device object for the class device (Physical Device Object)
//pContext: driver defined context, in this case our FDO (functional device object)
//return value: final status from the operation
NTSTATUS AmUsbPoSelfRequestCompletion( IN PDEVICE_OBJECT pDeviceObjectClass, IN UCHAR MinorFunction, IN POWER_STATE PowerState, IN PVOID pContext, IN PIO_STATUS_BLOCK pIoStatus )
{
	PDEVICE_OBJECT pDeviceObject = (PDEVICE_OBJECT)pContext;
	PDEVICE_EXTENSION pDeviceExtension = pDeviceObject->DeviceExtension;
	NTSTATUS Status = pIoStatus->Status;

	//we should not be in the midst of handling a system-generated power IRP
	AmUsbAssert( NULL == pDeviceExtension->pPowerIrp );

	//we only need to set the event if we're powering up;
	//no caller waits on power down complete
	if( (ULONG)PowerState.DeviceState < pDeviceExtension->PowerDownLevel )
		//trigger Self-requested power irp completed event;
		//the caller is waiting for completion
		KeSetEvent( &pDeviceExtension->evSelfRequestedPowerIrp, 1, FALSE );

	AmUsbDecIoCount( pDeviceObject );

	AmUsbDebugCond( DBGLVL_HIGH, !NT_SUCCESS( Status ), ( "exiting AmUsbPoSelfRequestCompletion() FAILED, Status = 0x%x\r\n", Status ) );
   
	return Status;
}

//This routine is called by BulkUsb_SelfSuspendOrActivate() to
//actually make the system request for a powerdown/up to PowerState.
//It first checks to see if we are already in Powerstate and immediately
//returns  SUCCESS with no further processing if so
//pDeviceObject: pointer to the device object
//PowerState: power state requested, e.g PowerDeviceD0.
//return value: final status from the operation
NTSTATUS AmUsbSelfRequestPowerIrp( IN PDEVICE_OBJECT pDeviceObject, IN POWER_STATE PowerState )
{
	NTSTATUS Status = STATUS_SUCCESS;
	PDEVICE_EXTENSION pDeviceExtension;
	PIRP pIrp = NULL;

	pDeviceExtension = pDeviceObject->DeviceExtension;

	//this should have been reset in completion routine
	AmUsbAssert( !pDeviceExtension->SelfPowerIrp );

	//nothing to do?
	if( pDeviceExtension->CurrentDevicePowerState == PowerState.DeviceState )
		return STATUS_SUCCESS;

	AmUsbDebug( DBGLVL_HIGH, ( "entering AmUsbSelfRequestPowerIrp(), will request power irp to state %s\r\n",
		AmUsbStringForDevState( PowerState.DeviceState ) ) );

	AmUsbIncIoCount( pDeviceObject );

	//flag we're handling a self-generated power irp
	pDeviceExtension->SelfPowerIrp = TRUE;

	//actually request the IRP
	Status = PoRequestPowerIrp( pDeviceExtension->pPhysicalDeviceObject, IRP_MN_SET_POWER,
			PowerState, AmUsbPoSelfRequestCompletion, pDeviceObject, NULL );
	if( Status == STATUS_PENDING )
	{ 
		//status pending is the return code we wanted

		//we only need to wait for completion if we're powering up
		if( (ULONG)PowerState.DeviceState < pDeviceExtension->PowerDownLevel )
		{
			NTSTATUS WaitStatus;

			WaitStatus = KeWaitForSingleObject( &pDeviceExtension->evSelfRequestedPowerIrp,
					Suspended, KernelMode, FALSE, NULL );
		}

		Status = STATUS_SUCCESS;

		//self power IRP is done
		pDeviceExtension->SelfPowerIrp = FALSE;

		AmUsbDebug( DBGLVL_HIGH, ( "AmUsbSelfRequestPowerIrp() SUCCESS, IRP 0x%x to state %s\r\n",
			pIrp, AmUsbStringForDevState(PowerState.DeviceState) ) );
	}
	else
	{
		//The return status was not STATUS_PENDING; any other codes must be considered in error here;
		//i.e., it is not possible to get a STATUS_SUCCESS  or any other non-error return from this call;
		AmUsbDebug( DBGLVL_HIGH, ( "AmUsbSelfRequestPowerIrp() to state %s FAILED, status = 0x%x\r\n",
			AmUsbStringForDevState( PowerState.DeviceState ), Status ) );
	}

	return Status;
}

//Called on BulkUsb_PnPAddDevice() to power down until needed (i.e., till a pipe is actually opened).
//Called on BulkUsb_Create() to power up device to D0 before opening 1st pipe.
//Called on BulkUsb_Close() to power down device if this is the last pipe.
//pDeviceObject: Pointer to the device object
//fSuspend: TRUE to Suspend, FALSE to acivate.
//return value:	SUCCESS, if the operation is not attemtped
//              final status from the operation, if the operation is attemtped
NTSTATUS AmUsbSelfSuspendOrActivate( IN PDEVICE_OBJECT pDeviceObject, IN BOOLEAN fSuspend )
{
	NTSTATUS Status = STATUS_SUCCESS;
	POWER_STATE PowerState;
	PDEVICE_EXTENSION pDeviceExtension;

	pDeviceExtension = pDeviceObject->DeviceExtension;

	AmUsbDebug( DBGLVL_MAXIMUM, ( "entering AmUsbSelfSuspendOrActivate(), fSuspend = %d\r\n", fSuspend ) );

	//can't accept request if:
	// 1) device is removed, 
	// 2) has never been started, 
	// 3) is stopped,
	// 4) has a remove request pending,
	// 5) has a stop device pending
	if( !AmUsbCanAcceptIoRequests( pDeviceObject ) )
	{
		Status = STATUS_DELETE_PENDING;
		AmUsbDebug( DBGLVL_MEDIUM, ( "ABORTING AmUsbSelfSuspendOrActivate()\r\n" ) );
		return Status;
	}

	//don't do anything if any System-generated Device PnP IRPs are pending
	if( pDeviceExtension->pPowerIrp != NULL )
	{
		AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbSelfSuspendOrActivate(), refusing on pending pDeviceExtension->PowerIrp 0x%x\r\n", pDeviceExtension->pPowerIrp ) );
		return Status;
	}

	//don't do anything if any self-generated Device Pnp irps are pending
	if( pDeviceExtension->SelfPowerIrp )
	{
		AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbSelfSuspendOrActivate(), refusing on pending pDeviceExtension->SelfPowerIrp\r\n" ) );
		return Status;
	}

	//don't auto-suspend if real driver is active
	if( fSuspend && pDeviceExtension->pDev != NULL )
	{
		AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbSelfSuspendOrActivate(), refusing to self-suspend, driver active\r\n" ) );
		return Status;
	}

	//don't auto-suspend if driver is opened by any apps
	if( fSuspend && pDeviceExtension->AppOpenCnt != 0 )
	{
		AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbSelfSuspendOrActivate(), refusing to self-suspend on AppOpenCnt %d\r\n", pDeviceExtension->AppOpenCnt ) );
		return Status;
	}

	//don't auto-activate if driver is not opened by any app
	if( !fSuspend && pDeviceExtension->AppOpenCnt == 0 )
	{
		AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbSelfSuspendOrActivate(), refusing to self-activate, AppOpenCnt 0\r\n" ) );
		return Status;
	}

	//don't do anything if registry CurrentControlSet\Services\BulkUsb\Parameters\PowerDownLevel
	//has been set to zero, PowerDeviceD0 ( 1 ), or a bogus high value
	if( pDeviceExtension->PowerDownLevel == PowerDeviceD0 || 
		pDeviceExtension->PowerDownLevel == PowerDeviceUnspecified ||
		pDeviceExtension->PowerDownLevel >= PowerDeviceMaximum )
	{
		AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbSelfSuspendOrActivate(), refusing on pDeviceExtension->PowerDownLevel == %d\r\n", pDeviceExtension->PowerDownLevel ) );
		return Status;
	}

	//suspend
	if( fSuspend )
		PowerState.DeviceState = pDeviceExtension->PowerDownLevel;
	//resume
	else
		PowerState.DeviceState = PowerDeviceD0; //power up all the way; we're probably just about to do some IO

	Status = AmUsbSelfRequestPowerIrp( pDeviceObject, PowerState );

	AmUsbDebug( DBGLVL_MAXIMUM, ( "AmUsbSelfSuspendOrActivate() status 0x%x on setting dev state %s\r\n", Status, AmUsbStringForDevState( PowerState.DeviceState ) ) );
	return Status;
}

//This routine generates an internal IRP from this driver to the PDO
//to obtain information on the Physical Device Object's capabilities.
//We are most interested in learning which system power states
//are to be mapped to which device power states for honoring IRP_MJ_SET_POWER Irps.
//This is a blocking call which waits for the IRP completion routine
//to set an event on finishing.
//pDeviceObject: Physical DeviceObject for this USB controller
//return value:	NTSTATUS value from the IoCallDriver() call
NTSTATUS AmUsbQueryCapabilities( IN PDEVICE_OBJECT pPdoDeviceObject, IN PDEVICE_CAPABILITIES pDeviceCapabilities )
{
	PIO_STACK_LOCATION pNextStack;
	PIRP pIrp;
	NTSTATUS Status;
	KEVENT evEvent;

	//This is a DDK-defined DBG-only macro that ASSERTS we are not running pageable code
	//at higher than APC_LEVEL.
	PAGED_CODE();

	//Build an IRP for us to generate an internal query request to the PDO
	pIrp = IoAllocateIrp( pPdoDeviceObject->StackSize, FALSE );

	if( !pIrp )
		return STATUS_INSUFFICIENT_RESOURCES;

	//IoGetNextIrpStackLocation gives a higher level driver access to the next-lower
	//driver's I/O stack location in an IRP so the caller can set it up for the lower driver
	pNextStack = IoGetNextIrpStackLocation( pIrp );
	AmUsbAssert( pNextStack != NULL );
	pNextStack->MajorFunction = IRP_MJ_PNP;
	pNextStack->MinorFunction = IRP_MN_QUERY_CAPABILITIES;

	//init an event to tell us when the completion routine's been called
	KeInitializeEvent( &evEvent, NotificationEvent, FALSE );

	//set a completion routine so it can signal our event when
	//the next lower driver is done with the Irp
	IoSetCompletionRoutine( pIrp, AmUsbIrpCompletionRoutine,
						   &evEvent, //pass the event as Context to completion routine
						   TRUE, TRUE, TRUE ); //invoke on success, on error and on cancellation of the IRP

	//set our pointer to the DEVICE_CAPABILITIES struct
	pNextStack->Parameters.DeviceCapabilities.Capabilities = pDeviceCapabilities;

	Status = IoCallDriver( pPdoDeviceObject, pIrp );

	AmUsbDebug( DBGLVL_MEDIUM,( "BulkUsb_QueryCapabilities(), Status from IoCallDriver to PCI = 0x%x\r\n", Status ) );

	if( Status == STATUS_PENDING )
	   //wait for IRP to complete
	   KeWaitForSingleObject( &evEvent, Suspended, KernelMode, FALSE, NULL );

	//failed? this is probably a bug
	AmUsbDebugCond( DBGLVL_DEFAULT, (!NT_SUCCESS( Status ) ), ( "AmUsbQueryCapabilities() FAILED\r\n" ) );

	IoFreeIrp( pIrp );

	return Status;
}
